/*
 The MIT License (MIT)

 Copyright (c) 2015 and ongoing Marcus Pöckelmann et al.

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/

// CATview - the Colored & Aligned Texts view
class CATview{
  constructor(_parent_id, _orientation, _x_inverted, _y_inverted){
    this.version = '3.beta.5.1';
    this.debug = false;
    if(this.debug)
      console.log('CATview.initialize');

    // • id of the parent container that will include CATview
    this.parent_id = _parent_id;

    // • the main svg element
    this.svg = null;
    this.width_svg = document.getElementById(this.parent_id).offsetWidth;
    this.height_svg = document.getElementById(this.parent_id).offsetHeight;

    // • CATviews orientation on the website
    this.orientation = ['left', 'right', 'bottom'].indexOf(_orientation) === -1 ? 'top' : _orientation;
    this.vertical = (this.orientation === 'left' || this.orientation === 'right');
    this.x_inverted = (typeof _x_inverted === 'undefined') ? false : _x_inverted;  // setting: default order of edges (false → left/top, true → right/bottom)
    this.y_inverted = (typeof _y_inverted === 'undefined') ? false : _y_inverted;  // setting: default order of witnesses (false → top/left, true → bottom/right)

    // • settings for dimensions
    this.margin = {left: 0, top: 0, right: 0, bottom: 0};
    this.space_for_egdes_axis = 24;       // reserved drawing size of the edges axis
    this.space_for_names_axis = 0;        // reserved drawing size of the name axis (calculated in respect to the actual names in .draw_svg)
    this.extra_space_for_names_axis = 4;  // some extra reserved space for the names axis
    this.space_for_tools = 22;            // reserved drawing size for tool icons
    this.space_for_other = 2;             // reserved size opposite to the edges axis
    this.content_padding = 5;             // extra padding of content area (e.g. left and right when horizontal)
    this.font_size_edges = 10;
    this.font_size_names = 14;

    // • colors of basic segments (rectangles)
    this.background = "#f8f8f8";
    this.scale_color = d3.scaleLinear().domain([0, 1]).range(["#C1C1E9", "#0000A3"]);
    this.rect_border = "stroke: #2f2f86;";
    this.equality_color = '#d9d9d9';

    // • initialization of toggable settings
    this.scroll_spy_enabled = true;     // display an orange bar (to indicate a scroll position)
    this.zebra_stripes_enabled = true;  // display gray lines in the background of every second witness
    this.extra_segments_enabled = true; // display additional rectangles in a dotted manner
    this.search_enabled = true;         // yellow highlighting of specific segments (e.g. for search results)
    this.search_mode = 'segment';       //    'segment' highlight search results as individual segments
                                        //    'edge'    highlight search results as full edges
    this.tools_enabled = true;          // display tool icons
    this.brush_enabled = false;         // display the interactive brushing tool
    this.rect_scaling_enabled = false;  // gives rectangles an individual size (defined by rect_scaling_data)
    this.rect_scaling_mode = 'default'; // defines the placement after scaling:
                                        //    'default'  rectangles are aligned on the bottom (left on vertical orientation)
                                        //    'middle'   rectangles are centered
                                        //    'inverted' rectangles are aligned on the top (right on vertical orientation)
    this.rect_linking_enabled = false;  // highlight linked segments on mouse hover
    this.drag_enabled = false;          // allows dragging of names to reorder them
    this.drag_mode = 'insert';          //    'insert' the dragged name is placed at target position, elements in between will be moved one position
                                        //    'swap'   switches places of dragged name and target position
    this.fadeout_enabled = true;        // fade out borders when zooming in
    this.equality_color_enabled = true; // display segments with value 0 in edges in a special color
    this.zoom_enabled = true;           // allows zooming via scrolling within the content area
    this.click_on_edge_enabled = true;  // react when a rectangle is clicked
    this.segment_texts_enabled = false; // text preview within segments/rectangles
    this.segment_texts_mode = 'l2r';    // writing direction of text within segments: 'l2r' | 'r2l'

    // • callbacks to add custom methods to certain events
    //  - first argument is always a reference to the CATview object
    this.click_on_edge_callback = null; // called when a rectangle was clicked (catview, index of edge)
    //  ‣ TODO ...
    this.brush_callback = null;
    this.brush_end_callback = null;
    this.drag_callback = null;          // called after a successful name swap
    this.show_rect_linking_callback = null;
    this.hide_rect_linking_callback = null;
    // this.dblclick_on_edge_callback = null; // unused yet

    // • internal variables
    this.scale_edges = null;          // scale to map the elements to the available pixels ([from, to])
    this.scale_names = null;          // map for names
    this.scale_edges_original = null; // copy of the initial map [0, n-1] (unzoomed)
    this.content = null;              // the content area for edges
    this.width_content = null;
    this.height_content = null;
    this.edges_axis = null;           // D3-element for the edges axis
    this.edge_names_box = null;       // DOM-element with the current name to display
    this.edge_names_enabled = true;   // set to false to disable the display the name of an edge on hovering
    this.rect_width = null;           // current dimensions for rectangles (depending on draw area and zoom)
    this.rect_height = null;
    this.rect_h_margin = null;
    this.rect_v_margin = null;
    this.rect_stroke = null;
    this.scroll_spy_pos = 1;          // default position of the scroll spy
    this.tools = [];                  // pairs of font-awesome-unicodes and callbacks

    // • variables for the brushing tool
    this.brush = null;
    this.brush_from_edge = -1;
    this.brush_to_edge = -1;
    this.brush_range = 1;
    this.brush_from_name = -1;
    this.brush_to_name = -1;
    this.brush_offset_from = -0.5;    // setting: offset for brushing box ...
    this.brush_offset_to = 0.5;       // ... in respect to the center of the selected rectangles (value on edges-axis)

    // • zooming variables
    this.zoom = null;       // the d3 object for zooming
    this.from = 1;          // positions of edges to display
    this.to = 1;
    this.from_pixel = 0;    // pixel-position in scale_edges_original of edges to display
    this.to_pixel = 0;
    this.scale = 1;         // current zooming factor
    this.translate = 0;     // current translation for zoomed-in excerpt
    this.zoom_button_clicked = false;
    this.zoom_step = 0.25;  // setting: steps by which to increase/decrease zoom in zoom_in/_out()

    // • data strcutures
    this.names = [];     // array with witnesses names (number of rows)
    this.edges = [];     // array with alignment edges (number of columns) - contains an array for each column (edge) with an entry for each column (witness), i.e.
                            //  -1: no segment for the witness
                            //  value within [0.0, 1.0]: segment, which will be colored according to this.scale_color
    this.search_results = [];  // array with column ids to be highlighted as search results
    this.extra_segments = [];  // additional segments that will be displayed in a faded out fashion (gray dotted without filling)
                                  //  entries are triples with [index of edge, index of witness, color value (within [0.0, 1.0])]
                                  // TODO sind hier noch im Format seg_wit, sollen aber analog zu data-segment-index zu wit_seg  
    this.edge_names = [];      // array of the edges' names, displayed on mouse over
    // linking of rectangles - highlights linked rectangles on mouse hover by color
    this.rect_linking_data = null;   // an array with groups of linked rectangles identified by their data-segment-index attribute, e.g.
    // [ ['0_19', '0_16'], ['1_25', '1_26', '1_27'] ]
    // (link two segments for the first witness; and three segments for the second)
    this.rect_scaling_data = null;       // same structure as edges - defines the size of rectangles
    this.segment_texts = null;      // same structure as edges - holds an preview text for the individual segment / rectangle


    // for text within segments
    this.segment_texts_min_scale = null;  // setting: defines the zooming factor at which texts will show up (ignored when set to null)
    this.segment_texts_min_width = 50;    // setting: defines the minimum rectangle width in px from which texts will show up (ignored when set to null)
    // TODO this.segment_texts_min_height → handle rect_scaling too

    // for scaling of rectangles
    this.rect_scaling_minimum = 0.2; // setting: defines the smallest possible size for a scaled rectangle

    // • drag'n'drop of names
    this.drag_names_order = null;    // the current order of names (might differ from the original order due to drag'n'drop)
    this.drag_order_name2pos = null; // array that maps: name index → position index
    this.drag_positions = [];
    this.drag_old_pos = null;
    this.drag_new_pos = null;
    this.drag_do_end = false;
    this.drag_threshold = 0.1;        // setting: TODO: wohl der Bereich um den target namen
  };

  // • setting and changing basic data (names and edges)
  //todo remove and set the identifiers of the witnesses
  set_names = function(_names){
    // todo add dummies or remove values in data structures
    this.names = _names;
    return true;
  };
  //add a name
  add_name = function(_name, _index){
    if(this.names == null)
      return false;
    let index = parseInt(Number(_index));
    if (isNaN(index) || index < 0 || index > this.names.length )
      index = this.names.length;

    // add the name
    this.names.splice(index, 0, _name);

    // add dummies to edges
    this.edges.forEach(function (edge) {
      edge.splice(index, 0, -1)
    });

    // refresh the search results
    this.search_results.forEach(function (edge) {
      for(let i = 0; i < edge[1].length; i++){
        if(edge[1][i] >= index)
          edge[1][i] += 1
      }
    });

    // refresh the extra segments
    this.extra_segments.forEach(function (segment) {
      if(segment[1] >= index)
        segment[1] += 1;
    });

    // refresh the data for rectangle scaling (add dummies as for edges)
    if(this.rect_scaling_data !== null){
      this.rect_scaling_data.forEach(function (edge) {
        edge.splice(index, 0, -1)
      });
    }

    // refresh the data for rectangle linking
    if(this.rect_linking_data !== null){
      this.rect_linking_data.forEach(function (group) {
        for(let i = 0; i < group.length; i++){
          let wit = parseInt(group[i].split('_')[0])
          if(wit >= index)
            group[i] = ( (wit + 1) + '_' + group[i].split('_')[1] )
        }
      });
    }

    // TODO may refresh drag_names_order ...

    return true;
  };
  //remove a witness and its values in all data structures
  remove_name = function(_index){
    let index = parseInt(Number(_index));
    if (isNaN(index) || this.names == null || index < 0 || index >= this.names.length )
      return false;

    this.names.splice(index, 1);

    // remove edges
    this.edges.forEach(function (edge) {
      edge.splice(index, 1)
    });

    // refresh the search results
    this.search_results.forEach(function (edge) {
      for(let i = edge[1].length - 1; i >= 0; i--){
        if(edge[1][i] > index)
          edge[1][i] -= 1
        else if (edge[1][i] == index)
          edge[1].splice(i, 1)
      }
    });

    // refresh the extra segments
    for(let i = this.extra_segments.length - 1; i >= 0; i--) {
      if(this.extra_segments[i][1] > index)
        this.extra_segments[i][1] -= 1
      else if (this.extra_segments[i][1] == index)
        this.extra_segments.splice(i, 1)
    }

    // refresh the data for rectangle scaling (remove within edges)
    if(this.rect_scaling_data !== null){
      this.rect_scaling_data.forEach(function (edge) {
        edge.splice(index, 1)
      });
    }

    // refresh the data for rectangle linking
    if(this.rect_linking_data !== null){
      this.rect_linking_data.forEach(function (group) {
        for(let i = 0; i < group.length; i++){
          let wit = parseInt(group[i].split('_')[0])
          if(wit > index)
            group[i] = ( (wit - 1 ) + '_'  + group[i].split('_')[0] )
          else if(wit == index)
            group.splice(i, 1)
        }
      });
    }

    // TODO may refresh drag_names_order ...

    return true;
  };
  // rename the witness at a specific index
  replace_name = function(_name, _index){
    let index = parseInt(Number(_index));
    if (isNaN(index) || this.names == null || index < 0 || index >= this.names.length )
      return false;

    this.names[index] = _name;

    // todo update edges

    return true;
  };
  // overwrite the alignment edges
  set_edges = function(_edges){
    this.edges = _edges;
    return true;
  };
  // add a single edge
  add_edge = function(_edge, _index){
    if(this.edges == null)
      return false;

    let index = parseInt(Number(_index));
    if (isNaN(index) || index < 0 || index > this.edges.length )
      index = this.edges.length;

    // handle malformed parameter
    while(edge.length > this.names.length)
      edge.pop();
    while(edge.length < this.names.length)
      edge.push(-1);

    this.edges.splice(index, 0, _edge);

    // refresh the search results
    for(let i = 0; i < this.search_results.length; i++){
      if(this.search_results[i][0] >= index)
        this.search_results[i][0] += 1
    }

    // refresh the extra segments
    this.extra_segments.forEach(function (segment) {
      if(segment[0] >= index)
        segment[0] += 1;
    });

    // refresh the data for rectangle scaling (add dummies as for edges)
    if(this.rect_scaling_data !== null){
      let scales = []
      for(let i = 0; i < edge.length; i++)
        if(edge[i] == -1)
          scales.push(-1);
        else
          scales.push(1);
      this.rect_scaling_data.splice(index, 0, scales);
    }

    // refresh the data for rectangle linking
    if(this.rect_linking_data !== null){
      this.rect_linking_data.forEach(function (group) {
        for(let i = 0; i < group.length; i++){
          let row = parseInt(group[i].split('_')[1])
          if(row >= index)
            group[i] = ( group[i].split('_')[0] + '_' + (row + 1))
        }
      });
    }

    // TODO may refresh available zooming level ...

    return true;
  };
  // remove a single edge
  remove_edge = function(_index){
    index = parseInt(Number(_index));
    if (isNaN(index) || this.edges == null || index < 0 || index >= this.edges.length )
      return false;

    this.edges.splice(index, 1);

    // refresh the search results
    for(let i = this.search_results.length - 1; i >= 0 ; i--){
      if(this.search_results[i][0] > index)
        this.search_results[i][0] -= 1
      else if (this.search_results[i][0] == index)
        this.search_results.splice(i, 1)
    }

    // refresh the extra segments
    for(let i = this.extra_segments.length - 1; i >= 0; i--) {
      if(this.extra_segments[i][0] > index)
        this.extra_segments[i][0] -= 1
      else if (this.extra_segments[i][0] == index)
        this.extra_segments.splice(i, 1)
    }

    // refresh the data for rectangle scaling (remove within edges)
    if(this.rect_scaling_data !== null){
      this.rect_scaling_data.forEach(function (edge) {
        edge.splice(index, 1)
      });
    }

    // refresh the data for rectangle scaling (add dummies as for edges)
    if(this.rect_scaling_data !== null){
      this.rect_scaling_data.splice(index, 1);
    }

    // refresh the data for rectangle linking
    if(this.rect_linking_data !== null){
      this.rect_linking_data.forEach(function (group) {
        for(let i = 0; i < group.length; i++){
          let row = parseInt(group[i].split('_')[1])
          if(row > index)
            group[i] = ( group[i].split('_')[0] + '_' +  (row - 1 ))
          else if(row == index)
            group.splice(i, 1)
        }
      });
    }

    // TODO may refresh available zooming level ...

    return true;
  };
  // replace a single edge
  replace_edge = function(_edge, _index, _extra){
    index = parseInt(Number(_index));
    if (isNaN(index) || this.edges == null || index < 0 || index >= this.edges.length )
      return false;

    this.edges[index] = _edge;

    if (_extra != null && !isNaN(index) && this.extra_segments != null && index > 0 && index < this.extra_segments.length )
      this.extra_segments[index] = _extra;

    return true;
  };
  // query new data from the server and call draw_svg
  server_update = function(_url, _callback){
    let catview = this;
    if(this.debug)
      console.log('CATview.server_update');

    let xmlhttp = new XMLHttpRequest();
    xmlhttp.onreadystatechange = function() {
      if(this.readyState === 4 && this.status === 200) {
        let data = JSON.parse(this.responseText);
        if(catview.debug)
          console.log(data);

        catview.names = data.names;
        // default order of names
        catview.drag_order_name2pos = [];
        for(let i = 0; i < catview.names.length; i++){
          catview.drag_order_name2pos.push(i);
        }
        catview.edges = data.edges;
        catview.edge_names = data.edge_names;
        catview.extra_segments = data.extra_segments;
        catview.rect_linking_data = data.rect_linking_data;
        catview.rect_scaling_data = data.rect_scaling;

        catview.draw_svg();
        if (_callback)
          _callback(catview);
      }
    };
    xmlhttp.open("GET", _url, true);
    xmlhttp.send();
    return true;
  };
  // slight change of the data
  // TOOD search_results, extra_segments, edge_names, rect_scaling, rect_linking_data
  // Auch bei geänderter Reihenfolge der names!
  data_change = function(_changes){
    if(this.debug)
      console.log('CATview.data_change');

    if(this.content) {
      // _changes should be an array with entries in the following form
      //  ['replace', col, edge]
      //  ['insert', col, edge]
      //  ['remove', col]
      // where 'col' is the index of the column and 'edge' the change information, i.e. an array with color values for each row at this col
      if (!(Object.prototype.toString.call( _changes ) === '[object Array]'))    // _changes could be a string if passed from Rails
        _changes = eval(_changes);
      if (Object.prototype.toString.call( _changes ) === '[object Array]') {
        for (let i = 0; i < _changes.length; i++) {
          // todo check the data types
          // todo search_results on replace (keyword needed)
          // todo remaining_edges (insert, remove)
          let type = _changes[i][0];
          let col = _changes[i][1];
          let edge = _changes[i][2][0]
          //console.log(type + ' ' + col);
          //console.log(_changes[i][2] );
          switch (type) {
            case 'replace':
              this.edges[col] = edge;
              // update the extra_segments
              for (let j = this.extra_segments.length - 1; j >= 0; j--) {
                if (this.extra_segments[j][0] === col && edge[this.extra_segments[j][1]] !== -1){
                  // extra segment will be deleted
                  // prior update corresponding rect linking, e.g. [ ['19_0', '16_0'], ['25_1', '26_1', '27_1'] ]
                  // TODO kann nicht mit Verschiebung umgehen!
                  // TODO update nach wit, edge
                  for (let jj = this.rect_linking_data.length - 1; jj >= 0; jj--) {
                    for (let k = this.rect_linking_data[jj].length - 1; k >= 0; k--) {
                      if (this.rect_linking_data[jj][k] === '' + col + '_' + this.extra_segments[j][1] ) {
                        this.rect_linking_data[jj].splice(k, 1);
                      }
                    }
                    if(this.rect_linking_data[jj].length <= 1)
                      this.rect_linking_data.splice(jj, 1);
                  }
                  // and delete the extra segment
                  this.extra_segments.splice(j, 1);

                }
              }
              // TODO update the rect_scaling
              break;
            case 'insert':
              this.edges.splice(col, 0, edge);
              // update the search results
              for (let j = 0; j < this.search_results.length; j++) {
                if (this.search_results[j][0] > col)
                  this.search_results[j][0] += 1;
              }
              // update the extra_segments
              for (let j = 0; j < this.extra_segments.length; j++) {
                if (this.extra_segments[j][0] > col)
                  this.extra_segments[j][0] += 1;
              }
              // update the rect_linking_data, e.g. [ ['19_0', '16_0'], ['25_1', '26_1', '27_1'] ]
              // TODO update nach wit, edge
              for (let j = 0; j < this.rect_linking_data.length; j++) {
                for (let k = 0; k < this.rect_linking_data[j].length; k++) {
                  let pos = this.rect_linking_data[j][k].split('_') // [col, wit]
                  let _col = parseInt(pos[0])
                  if(_col > col){
                    this.rect_linking_data[j][k] = '' + (_col + 1 ) + '_' + pos[1]
                  }
                }
              }
              // update the edge_names
              this.edge_names.splice(col, 0, '');
              // TODO update the rect_scaling
              // TODO update the brush if displayed
              break;
            case 'remove':
              this.edges.splice(col, 1);
              // update the search results
              for (let j = this.search_results.length - 1; j >= 0; j--) {
                if (this.search_results[j][0] > col)
                  this.search_results[j][0] -= 1;
                else if (this.search_results[j][0] === col)
                  this.search_results.splice(j, 1);
              }
              // update the extra_segments
              for (let j = this.extra_segments.length - 1; j >= 0; j--) {
                if (this.extra_segments[j][0] > col)
                  this.extra_segments[j][0] -= 1;
                else if (this.extra_segments[j][0] === col)
                  this.extra_segments.splice(j, 1);
              }
              // update the rect_linking_data, e.g. [ ['19_0', '16_0'], ['25_1', '26_1', '27_1'] ]
              // TODO update nach wit, edge
              for (let j = this.rect_linking_data.length - 1; j >= 0; j--) {
                for (let k = this.rect_linking_data[j].length - 1; k >= 0; k--) {
                  let pos = this.rect_linking_data[j][k].split('_') // [col, wit]
                  let _col = parseInt(pos[0])
                  if(_col > col)
                    this.rect_linking_data[j][k] = '' + (_col - 1 ) + '_' + pos[1]
                  else if (_col === col) {
                    this.rect_linking_data[j].splice(k, 1);
                  }
                }
                if(this.rect_linking_data[j].length <= 1)
                  this.rect_linking_data.splice(j, 1);
              }
              // update the edge_names
              this.edge_names.splice(col, 1);
              // TODO update the rect_scaling
              // TODO update the brush if displayed
              break;
            default:
              console.log('CATview.data_change: unknown kind of change (' + _changes[i][0] + ')');
              break;
          }
        }

        this.set_scale_edges(0, this.edges.length + 1);
        this.scale_edges_original = this.scale_edges.copy();
        this.refresh_content(this.from, this.to);
        return true;
      }
      else
        return false;
    }
    else
      return false;
  };

  // • define and draw the basic elements
  // draw the initial svg that contains CATview and call refresh_content
  draw_svg = function () {
    if(this.debug)
      console.log('CATview.draw_svg');

    // remove previous content and add the svg-element to the parent
    document.getElementById(this.parent_id).innerHTML = '';
    this.svg = d3.select('#' + this.parent_id).insert("svg", ":first-child").attr('width', this.width_svg).attr('height', this.height_svg);

    // determine the necessary padding before the names-axis in respect to the names
    let names_offset = 0;
    if(this.names.length > 0){
      this.svg.append("g")
      .attr('id', this.parent_id + '_prerendered_names')
      .selectAll("text")
      .data(this.names)
      .enter().append("text")
      .text(function (d) {return d})
      .attr('font-size', this.font_size_names);
      // get the dimension of the longest text and add 2 pixels padding
      let names = document.querySelectorAll('#' + this.parent_id + '_prerendered_names text');
      for(let i = 0; i < names.length; i++){
        if(names_offset < names[i].getComputedTextLength())
          names_offset = names[i].getComputedTextLength();
      }
      this.space_for_names_axis = names_offset + this.extra_space_for_names_axis;
      //// than remove this temporary elements
      let element = document.getElementById(this.parent_id + '_prerendered_names');
      element.parentNode.removeChild(element);
    }

    // initialize margins as well as width and height for the content
    this.set_margins();

    // initialize the scales
    this.set_scale_edges(0, this.edges.length + 1)
    this.scale_edges_original = this.scale_edges.copy();
    this.set_scale_names();

    // add some zebra stripes
    this.svg.append("g").attr("class", "zebra");
    this.draw_zebra_stripes();

    // add the names-axis
    if(this.drag_enabled)
      this.svg.append("g").attr("class", "axis names-axis draggable");
    else
      this.svg.append("g").attr("class", "axis names-axis");
    this.draw_names_axis();

    // add the group of tool icons
    this.svg.append("g").attr("class", "tool-icons");
    this.draw_tools();

    // add a group for the content and its zoom behavior
    this.content = this.svg.append("g")
    .attr("class", "content")
    .attr("transform", "translate(" +
      (this.margin.left + (this.vertical ? 0 : this.content_padding)) + "," +
      (this.margin.top + (this.vertical ? this.content_padding : 0)) + ")");

    // add the zooming behavior
    this.zoom = d3.zoom().on("zoom", this.zooming.bind(this)); // bind the CATview object to make it directly available within zooming()
    this.set_max_zoom();
    this.content.call(this.zoom)
    .on("dblclick.zoom", null); // disable zoom on double click

    // the zoom needs an element over the contents' complete size
    this.content.append('rect')
    .attr('width', this.width_content)
    .attr('height', this.height_content)
    .attr('style', 'fill-opacity: 0.0;');

    // add the edges-axis
    this.content.append("g").attr("class", "axis edges-axis");
    this.draw_edges_axis();

    // add containers for different groups of content
    this.content.append("g").attr("class", "search_results_background");
    this.content.append("g").attr("class", "alignment");

    // set default interval of data to be shown and draw the content
    this.from_pixel = 0;
    this.to_pixel = this.vertical === true ? this.height_content : this.width_content;
    this.refresh_content(1, this.edges.length);

    // append the brush, if it was enabled prior the call of draw_svg
    this.draw_brush();

    // append box to display an edges name on hover
    this.edge_names_box = d3.select("#"+this.parent_id).append("div").attr("class", "edge-name" + " edge-name-" + this.orientation);
  };

  // method to set margins and content width/height according to the current orientation of CATview and its axis
  set_margins = function(){
    if(this.debug)
      console.log('CATview.set_margins');

    if(this.vertical){
      // handle place of names axis and tool icons in respect to the edges axis direction for vertical orientation of CATview
      this.margin.top = this.x_inverted ? this.space_for_tools : this.space_for_names_axis;
      this.margin.bottom = this.x_inverted ? this.space_for_names_axis : this.space_for_tools;
      // handle place of the edges axis according to the (left or right) orientation of CATview
      this.margin.left = this.orientation === 'left' ? this.space_for_other : this.space_for_egdes_axis;
      this.margin.right = this.orientation === 'left' ? this.space_for_egdes_axis : this.space_for_other;
    } else {
      // handle place of the edges axis according to the (top or bottom) orientation of CATview
      this.margin.top = this.orientation === 'top' ? this.space_for_other : this.space_for_egdes_axis;
      this.margin.bottom = this.orientation === 'top' ? this.space_for_egdes_axis : this.space_for_other;
      // handle place of names axis and tool icons in respect to the edges axis direction for horizontal orientation of CATview
      this.margin.left = this.x_inverted ? this.space_for_tools : this.space_for_names_axis;
      this.margin.right = this.x_inverted ? this.space_for_names_axis : this.space_for_tools;
    }

    // set the available width and height for the content in respect to the margin
    this.width_content   = this.width_svg - this.margin.left - this.margin.right;
    this.height_content  = this.height_svg - this.margin.top - this.margin.bottom;
    // add some padding for the content
    if(this.vertical)
      this.height_content -= 2 * this.content_padding;
    else
      this.width_content -= 2 * this.content_padding;
  };

  // draws the axis for the names of text witnesses and the edges of aligned segments
  draw_names_axis = function(){
    let catview = this;
    if(this.debug)
      console.log('CATview.draw_names_axis');

    let names_axis = this.svg.select(".names-axis");
    names_axis.selectAll('*').remove(); // remove the current content

    // place the names axis
    if(this.x_inverted)
      if(this.vertical)
        names_axis.attr("transform", "translate(" + this.margin.left + ", " + (this.height_svg - this.margin.bottom) + ")");
      else
        names_axis.attr("transform", "translate(" + (this.width_svg - this.margin.right) + ", " + this.margin.top + ")");
    else
      names_axis.attr("transform", "translate(" + this.margin.left + ", " + this.margin.top + ")");

    // default names order (may not initialised yet or number of names are not up-to-date)
    if(this.drag_order_name2pos === null || this.drag_order_name2pos.length != this.names.length){
      this.drag_order_name2pos = [];
      for(let i = 0; i < this.names.length; i++){
        this.drag_order_name2pos.push(i);
      }
    }

    // add the names
    let anchor = 'end';
    if(this.vertical){
      if((this.orientation === 'left' && this.x_inverted) || (this.orientation === 'right' && !this.x_inverted))
        anchor = 'start';
      names_axis.selectAll("text")
      .data(this.names)
      .enter().append("text")
      .text(function (d) {return d})
      .attr("x", function(d, i) {return catview.scale_names(catview.drag_order_name2pos[i])})
      .attr("original", function(d, i) {return i})
      .attr("current", function(d, i) {return catview.drag_order_name2pos[i]})
      .attr("style", "text-anchor: " + anchor + ";")
      .attr('font-size', this.font_size_names)
      .attr("transform", function(d, i) {
        // rotate for 90 degrees than move the names a bit to the right to make them centered
        return 'rotate(' + (catview.orientation === 'left' ? 90 : -90) + ', ' + catview.scale_names(catview.drag_order_name2pos[i]) + ', 0)translate(0,' + (catview.font_size_names/2) + ')';
      });
    } else {
      if(this.x_inverted)
        anchor = 'start';
      names_axis.selectAll("text")
      .data(this.names)
      .enter().append("text")
      .text(function (d) {return d})
      .attr("y", function(d, i) {return catview.scale_names(catview.drag_order_name2pos[i]) + 5 }) // a little offset due to font size
      .attr("original", function(d, i) {return i})
      .attr("current", function(d, i) {return catview.drag_order_name2pos[i]})
      .attr("style", "text-anchor: " + anchor + ";")
      .attr('font-size', this.font_size_names);
    }

    // enable drag'n'drop of names (and their associated edges)
    if(this.drag_enabled){
      // which axis the draggable names are placed
      // get the valid positions from dragging names
      this.drag_positions = [];
      d3.selectAll("#"+this.parent_id+" g.axis.names-axis>text").each(function(d,i) {
        catview.drag_positions.push( parseFloat(d3.select(this).attr((catview.vertical? "x" : "y"))));
      });
      // and sort them in ascending order
      this.drag_positions.sort(function(a, b){ return a-b; });

      // define the drag events
      var drag_handler = d3.drag()
      .on("start", function(){
        let drag_object = d3.select(this);
        catview.drag_old_pos = parseInt(drag_object.attr('current'));

        d3.select("#"+catview.parent_id+" g.axis.names-axis")
        .append("text")
        .text(drag_object.text())
        .attr("id", "ghost")
        .style("visibility", "hidden")
        .style("opacity", 0.5)
        .style("text-anchor", "middle")
        .attr("transform", "rotate("+(catview.orientation === "left"? "90" : catview.orientation === "right"? "-90" : "0") + ")" );

      })
      .on("drag", function(d){
        // track the current grabbed name label
        d3.select("#ghost")
        .style("visibility", "visible")
        .style("cursor", "grabbing")
        .attr("x", d3.event.x)
        .attr("y", d3.event.y);

        if (!catview.vertical){
          d3.select("#ghost")
          .style("visibility", "visible")
          .style("cursor", "grabbing")
          .attr("x", d3.event.x)
          .attr("y", d3.event.y);
        }
        else {
          if (catview.orientation === "left"){
            d3.select("#ghost")
            .style("visibility", "visible")
            .style("cursor", "grabbing")
            .attr("x", d3.event.y)
            .attr("y", (-1)*d3.event.x);
          }
          else {
            d3.select("#ghost")
            .style("visibility", "visible")
            .style("cursor", "grabbing")
            .attr("x", (-1)*d3.event.y)
            .attr("y", d3.event.x);
          }
        }

        // if current mouse pos differs enough from valid pos, then choose
        // the nearest neighbour and switch positions with it if there is a neighbour
        if (Math.abs((catview.vertical? d3.event.x : d3.event.y) - catview.drag_positions[catview.drag_old_pos]) >
          catview.drag_threshold * Math.abs(catview.drag_positions[0]-catview.drag_positions[1])){
          // only when the attributes here have been updated, we should do a drag-end event.
          // if not, this results in an error when double clicking on an element, after a first
          // drag has been applied
          catview.drag_do_end = true;

          // we don't need the direction, we just need to identify the position of the dragged element
          // with the one with the closest coordinates to the mouse cursor

          // select the element with the closest y to current mouse position
          // for this calculate all differences of the valid positions to the mouse cursor
          // and take the index with lowest distance.
          let distances = catview.drag_positions.map(function(el){return Math.abs(el-(catview.vertical? d3.event.x : d3.event.y));});
          // get the index with minimal distance
          let minimum = Math.min(...distances); // ... → translates the array to a list of values
          let idx_min = distances.indexOf(minimum); // just returns first index that matches argument

          // now we can estimate the targets position
          catview.drag_new_pos = idx_min;
        }
      })
      .on("end", function(d){
        // only swap on end of dragging!
        // swap the dragged elements and the target elements position:
        d3.select("#ghost").remove();

        if(!catview.drag_do_end)
          return;

        let old_pos = catview.drag_old_pos;
        let new_pos = catview.drag_new_pos;
        let objects = {};
        // get all name objects
        for(let pos = 0; pos < catview.names.length; pos++){
          objects[pos] = d3.select("#"+catview.parent_id+" g.axis.names-axis>text[current=\"" + pos + "\"]");
        }

        // collects all information necessary to move the names (and their associated segments)
        let names_to_move = [];
        // add the dragged name
        names_to_move.push({
          'object': objects[old_pos],
          'from': old_pos,
          'to': new_pos,
          'new_position': catview.drag_positions[new_pos], // move_pos[move_pos.length-1]
          'transform': objects[new_pos].attr("transform") // of name on target position
        });
        // add other names depending on the drag mode:
        if(catview.drag_mode === 'swap'){
          // for swap mode → add the name at the target position
          names_to_move.push({
            'object': objects[new_pos],
            'from': new_pos,
            'to': old_pos,
            'new_position': catview.drag_positions[old_pos],
            'transform': objects[old_pos].attr("transform") // of name on target position
          });
        }
        else{
          // for insert mode = a ring swap → add the names between the old and new position of the dragged name
          if(old_pos < new_pos){
            for (let pos = new_pos; pos > old_pos; pos--){
              names_to_move.push({
                'object': objects[pos],
                'from': pos,
                'to': pos - 1,
                'new_position': catview.drag_positions[pos - 1],
                'transform': objects[pos - 1].attr("transform")
              });
            }
          }
          else{ // new_pos < old_pos
            for (let pos = new_pos; pos < old_pos; pos++){
              names_to_move.push({
                'object': objects[pos],
                'from': pos,
                'to': pos + 1,
                'new_position': catview.drag_positions[pos + 1],
                'transform': objects[pos + 1].attr("transform")
              });
            }
          }

        }

        // perform the movment of the names and their associated segments
        names_to_move.forEach(function(name, i){
          // move the name
          name.object.transition()
          .attr((catview.vertical? "x" : "y"), name.new_position)
          .attr("transform", name.transform); // add transformation of the text (rotation in vertical orientation)

          // remember the new current position
          name.object.attr('current', name.to);
          // and update the names order
          let original = parseInt(name.object.attr('original'));
          catview.drag_order_name2pos[original] = name.to;

          // move the associated segments
          let segments = d3.selectAll("#" + cat.parent_id + " .segment[data-segment-index^=\""+original+"_\"]");

          // set the x or y attribute to the one at the position before (in move_pos)
          let add_to_source = catview.drag_positions[name.to] - catview.drag_positions[name.from];
          // segments can differ in height and therefore y-value (in vertical orientation)
          // so we need to add the same amount instead of setting the y value the same for all
          segments.each( function(){
            //let translation = d3.transform(d3.select(this).attr("transform")).translation
            // parse transform attribute (as d3.transform was removed in d3.v4)
            let transform = d3.select(this).attr("transform") // as
            let translation = transform.substring(transform.indexOf('(') + 1, transform.indexOf(')')).split(',');
            if(catview.vertical === true)
              translation = "translate(" + (parseFloat(translation[0]) + add_to_source) + ", 0)";
            else
              translation = "translate(0, " + (parseFloat(translation[1]) + add_to_source) + ")";
            d3.select(this)
            .transition()
            .attr('transform', translation);
          });
        });

        // drag finished
        catview.drag_do_end = false;

        if (catview.drag_callback !== null)
          return catview.drag_callback(catview, old_pos, new_pos);
      });

      // select the name labels and listen for drag-events
      let inner_texts = d3.selectAll("#"+catview.parent_id+" g.axis.names-axis>text");
      drag_handler(inner_texts);
    }
    return true;
  };
  draw_edges_axis = function(){
    if(this.debug)
      console.log('CATview.draw_edges_axis');

    let axis = this.svg.select(".edges-axis");
    axis.selectAll('*').remove(); // remove the current content

    let pos_x = 0;
    let pos_y = 0;
    switch(this.orientation) {
      case 'left':
        this.edges_axis = d3.axisRight(this.scale_edges);
        pos_x = this.width_content;
        break;
      case 'right':
        this.edges_axis = d3.axisLeft(this.scale_edges);
        break;
      case 'bottom':
        this.edges_axis = d3.axisTop(this.scale_edges);
        break;
      default:
        // numbers of edges at the bottom for top orientation of CATview
        this.edges_axis = d3.axisBottom(this.scale_edges);
        pos_y = this.height_content;
    }
    axis
    .attr("transform", "translate(" + pos_x + "," + pos_y + ")")
    .call(this.edges_axis);
    return true;
  };
  // methods to set scales according to the current orientation of CATview and its axis
  set_scale_names = function(){
    if(this.debug)
      console.log('CATview.set_scale_names');

    if(this.vertical){
      // vertical orientation (left, right)
      if(this.y_inverted){
        // inverted names axis (starting right)
        this.scale_names = d3.scaleLinear().domain([-0.5, this.names.length - 0.5]).range([this.width_content, 0]);
      } else {
        // normal names axis (starting left)
        this.scale_names = d3.scaleLinear().domain([-0.5, this.names.length - 0.5]).range([0, this.width_content]);
      }
    } else {
      // horizontal orientation (top, bottom)
      if(this.y_inverted){
        // inverted names axis (starting bottom)
        this.scale_names = d3.scaleLinear().domain([-0.5, this.names.length - 0.5]).range([this.height_content, 0]);
      } else {
        // normal names axis (starting top)
        this.scale_names = d3.scaleLinear().domain([-0.5, this.names.length - 0.5]).range([0, this.height_content]);
      }
    }
  };
  set_scale_edges = function(_from, _to){
    if(this.debug)
      console.log('CATview.set_scale_edges');

    if(this.vertical){
      // vertical orientation (left, right)
      if(this.x_inverted){
        // inverted edges axis (starting bottom)
        this.scale_edges = d3.scaleLinear().domain([_from+0.5, _to-0.5]).range([this.height_content, 0]);
      } else {
        // normal edges axis (starting top)
        this.scale_edges = d3.scaleLinear().domain([_from+0.5, _to-0.5]).range([0, this.height_content]);
      }
    } else {
      // horizontal orientation (top, bottom)
      if(this.x_inverted){
        // inverted edges axis (starting right)
        this.scale_edges = d3.scaleLinear().domain([_from+0.5, _to-0.5]).range([this.width_content, 0]);
      } else {
        // normal edges axis (starting left)
        this.scale_edges = d3.scaleLinear().domain([_from+0.5, _to-0.5]).range([0, this.width_content]);
      }
    }
  };
  // methods to reverse the order of text witnesses and edges
  invert_names_axis = function(){
    if(this.debug)
      console.log('CATview.invert_names_axis');

    let name2pos = this.get_names_order(); //  map: name index → position index
    let pos2name = Array(name2pos.length);    //  create map: position index → name index
    for(let i = 0; i < name2pos.length; i++)
      pos2name[name2pos[i]] = i;
    pos2name.reverse();                       // invert the positions
    for(let i = 0; i < pos2name.length; i++)  // remap to: name index → position index
      name2pos[pos2name[i]] = i;

    return this.set_names_order(name2pos);
  };
  invert_edges_axis = function(){
    if(this.debug)
      console.log('CATview.invert_edges_axis');

    this.x_inverted = !this.x_inverted;

    if(this.content){
      this.set_margins();
      this.set_scale_edges(0, this.edges.length + 1);
      this.scale_edges_original = this.scale_edges.copy();
      // relocate the content container
      this.svg.select(".content").attr("transform", "translate(" +
        (this.margin.left + (this.vertical ? 0 : this.content_padding)) + "," +
        (this.margin.top + (this.vertical ? this.content_padding : 0)) + ")");
      this.draw_zebra_stripes();
      this.draw_edges_axis();
      this.draw_names_axis();
      this.draw_tools();
      this.refresh_content(this.from, this.to);
    }
    else
      return false;
  };

  // define the interval of edges and refresh the content afterwards
  refresh_content = function(_from, _to){
    if(this.debug)
      console.log('CATview.refresh_content(' + _from + ', ' + _to + ')');

    if(_from === undefined && this.from === null)
      _from = 1;
    if(_to === undefined && this.to === null)
      _to = this.edges.length;

    if(this.content){
      if (_from !== null && _from !== undefined) this.from = parseInt(_from);
      if (this.from < 1) this.from = 1;
      if (_to !== null &&_to !== undefined) this.to = parseInt(_to);
      if (this.to > this.edges.length) this.to = this.edges.length;
      // refresh the edges-scale (with one bin offset)
      this.set_scale_edges(Math.max(this.from - 1, 0), Math.min(this.to + 1, this.edges.length + 1));

      this.draw_alignment();
      this.draw_extra_segments();
      this.draw_search_results();
      this.draw_segment_texts();
      if(this.scroll_spy_enabled)
        this.draw_scroll_spy();
    }
    else
      return false;
  };

  // draws the alignment as colored rectangles as well as the edges-axis
  draw_alignment = function (){
    let catview = this;
    if(this.debug)
      console.log('CATview.draw_alignment: ' + this.from + " - " + this.to );


    if(this.content){
      // remove the edges of the previous visualization
      this.content.select(".alignment").selectAll("g.row_witness").remove();
      // get all edges within the given interval
      let edges_temp = this.edges.map(function(d, i) { return [d, i + 1];}).filter(function(d) { return catview.from <= d[1] && d[1] <= catview.to; });

      // get the max. width and height that is available for a segment
      let segment_width = Math.floor(this.width_content/(this.vertical === true ? this.names.length : edges_temp.length));
      if(segment_width < 1) segment_width = 1;
      let segment_height = Math.floor(this.height_content/(this.vertical === true ? edges_temp.length : this.names.length));
      if(segment_height < 1) segment_height = 1;

      // set width, height, margin and stroke for a rectangle
      this.rect_h_margin = Math.floor(segment_width/5);
      this.rect_width = segment_width - this.rect_h_margin;
      this.rect_v_margin = Math.floor(segment_height/5);
      this.rect_height = segment_height - this.rect_v_margin;
      this.rect_stroke = "stroke-width: " + ((segment_width >= 15 && segment_height >= 15) ? ((segment_width >= 30 && segment_height >= 30) ? 2 : 1) : 0) + ";";

      // create a grouping row for each edge
      let rows = this.content.select(".alignment").selectAll("g.row_witness")
      .data(edges_temp)
      .enter().append("g")
      .attr("class", "row_witness")
      .attr("transform", function(d) { return "translate(" +
        (catview.vertical === true ? (-catview.rect_width/2 + ", " + (catview.scale_edges(parseInt(d[1]))-(catview.rect_height/2))) :
          (catview.scale_edges(parseInt(d[1]))-(catview.rect_width/2)) + ", " + (-catview.rect_height/2))
        + ")"; })
      .on('mouseenter', function(d, i){catview.show_edge_name(i)})
      .on('mouseleave', function(){ catview.hide_edge_name()})
      .on("click", function(d){
        if(catview.click_on_edge_enabled && catview.click_on_edge_callback !== null)
        catview.click_on_edge_callback(catview, d[1]);
      });

      // draw the rects for the witnesses in each row
      let segments = rows.selectAll("g.segment")
      .data(function(d) { return d[0].map(function(d2, j) {return [j, (d[1] - 1), d2];}).filter( function(d2) { return d2[2] != "-1"; } ); })
      .enter()
      .append("g")
      .attr("class", "segment")
      .attr("data-segment-index", function(d) { return d[0] + "_" + d[1];})
      .attr("transform", function(d) {
        if(catview.vertical === true)
          return "translate(" + catview.scaled_position(d[1], d[0]) + ", 0)";
        else
          return "translate(0, " + catview.scaled_position(d[1], d[0]) + ")";})
      .on("mouseenter", function(d){ catview.show_rect_linking(d[0]+"_"+d[1]); })
      .on("mouseleave", function(d){ catview.hide_rect_linking(); })
      .attr('cursor', 'pointer')

      segments.append("rect")
      .attr("class", "segment-rect")
      .attr("width", function(d){ return catview.scaled_rect_width(d[1], d[0]); })
      .attr("height", function(d){ return catview.scaled_rect_height(d[1], d[0]); })
      //.attr("ry", rect_corner)      // rounded Corners
      .attr("style", function(d) {
        return catview.rect_stroke + catview.rect_border + "fill: " +
          (parseFloat(d[2]) === 0.0 && catview.equality_color_enabled ? catview.equality_color : catview.scale_color(parseFloat(d[2])));
      })

      segments.append("text")
      .attr("class", "segment-text")
      .attr("font-family", "Courier New")             // mono font

      // update the edges-axis
      this.edges_axis.scale(this.scale_edges);
      this.content.select(".edges-axis").call(this.edges_axis);
      this.content.select(".edges-axis").selectAll("text")
      .attr("style", "text-anchor: middle;")
      .attr("transform", function() {
        // rotate for 90 degrees than move the positions a bit to make them centered
        let transform = '';
        switch (catview.orientation)
        {
          case "left": transform = 'translate(3, 0)'; break;
          case "right": transform = 'translate(-3, 0)'; break;
          default: break;
        }
        return transform;
      })
      .attr('font-size', this.font_size_edges);


      // define gradients
      d3.select("defs#"+this.parent_id+"_gradient_definition").remove();

      this.content.select(".alignment")
      .append("defs")
      .attr("id", this.parent_id+"_gradient_definition")
      .append("linearGradient")
      .attr("id", this.parent_id+"_gradient_left")
      .attr("x1", "0" ) //x1="0" x2="0" y1="0" y2="1"
      .attr("x2", this.vertical === true ? "0" : "1")
      .attr("y1", "0")
      .attr("y2", this.vertical === true ? "1" : "0")
      .append("stop")
      .attr("offset", "20%")
      .attr("stop-color", this.background) // "rgba(248,248,248,1)") // "rgba(0,0,248,1)"
      .attr("stop-opacity", 1.0);
      //.attr("stop-color", "rgba(248,248,248,1)") // "rgba(0,0,248,1)"
      this.content.select(".alignment").select("defs").select("linearGradient#"+this.parent_id+"_gradient_left")
      .append("stop")
      .attr("offset", "80%")
      .attr("stop-color", this.background) // "rgba(248,248,248,1)") // "rgba(0,0,248,1)"
      .attr("stop-opacity", 0.0);
      //.attr("stop-color", "rgba(248,248,248,0)"); // hier 0 statt 1 für Transparenz

      // right gradient
      this.content.select(".alignment").select("defs#"+this.parent_id+"_gradient_definition")
      .append("linearGradient")
      .attr("id", this.parent_id+"_gradient_right")
      .attr("x1", "0") //x1="0" x2="0" y1="0" y2="1"
      .attr("x2", this.vertical === true ? "0" : "1")
      .attr("y1", "0")
      .attr("y2", this.vertical === true ? "1" : "0")
      .append("stop")
      .attr("offset", "20%")
      .attr("stop-color", this.background) // "rgba(248,248,248,1)") // "rgba(0,0,248,1)"
      .attr("stop-opacity", 0.0);
      //.attr("stop-color", "rgba(248,248,248,0)")
      this.content.select(".alignment").select("defs").select("linearGradient#"+this.parent_id+"_gradient_right")
      .append("stop")
      .attr("offset", "80%")
      .attr("stop-color", this.background) // "rgba(248,248,248,1)") // "rgba(0,0,248,1)"
      .attr("stop-opacity", 1.0);
      //.attr("stop-color", "rgba(248,248,248,1)") // "rgba(0,0,248,1)" (blue) , "rgba(248,248,248,1)" (transparent)

      // also draw faded borders if enabled
      this.draw_fadeout();
    } // if content
    else
      return false;
  };


  // • additional elements
  // ‣ draws the zebra stripes (gray lines in the background for every second witness)
  draw_zebra_stripes = function(){
    let catview = this;
    if(this.debug)
      console.log('CATview.draw_zebra_lines');

    let zebra = this.svg.select(".zebra");
    zebra.selectAll('*').remove();  // remove the current content

    if(!this.zebra_stripes_enabled)
      return false;

    let width, height;
    if(this.vertical){
      height = this.height_svg - this.space_for_tools - 2;
      width = Math.floor(this.width_content / this.names.length * 0.5)

    } else {
      width = this.width_svg - this.space_for_tools - 2;
      height = Math.floor(this.height_content / this.names.length * 0.5);
    }

    let offset = this.x_inverted ? this.space_for_tools : 0

    zebra.selectAll("rect")
    .data(this.names)
    .enter().append("rect")
    .attr("class", "rect-zebra")
    .attr("width", width )
    .attr("height", height)
    .attr('x', this.vertical === true ? function(d, i) {return catview.scale_names(i) - width / 2.0 + catview.margin.left} : offset)
    .attr('y', this.vertical === true ? offset : function(d, i) {return catview.scale_names(i) - height / 2.0 + catview.margin.top})
    .attr("fill", function(d,i) { return i%2===1 ? "#e8e8e8":"transparent";})
  };
  toggle_zebra_stripes = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_zebra_stripes(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.zebra_stripes_enabled = _enable;
    else
      this.zebra_stripes_enabled = !this.zebra_stripes_enabled;
    return this.draw_zebra_stripes();
  };

  // ‣ draws the scroll spy as orange bar
  draw_scroll_spy = function(_index){
    if(this.debug)
      console.log('CATview.draw_scroll_spy(' + _index + ')');

    if (arguments.length === 1)
      this.scroll_spy_pos = parseInt(_index) + 1;

    if(this.content){
      this.content.selectAll("rect.scroll_spy").remove();
      if(this.scroll_spy_enabled){
        if(this.from <= this.scroll_spy_pos && this.scroll_spy_pos <= this.to)
        {
          if(this.vertical === true){
            let height = this.rect_height + 2 * (this.rect_v_margin/2);
            this.content.insert("rect", ".edges-axis")
            .attr("class", "scroll_spy")
            .attr("width", this.width_content)
            .attr("height", height)
            .attr("x", 0)
            .attr("y", this.scale_edges(this.scroll_spy_pos) - (height/2));
          }
          else{
            let width = this.rect_width + 2 * (this.rect_h_margin/2);
            this.content.insert("rect", ".edges-axis")
            .attr("class", "scroll_spy")
            .attr("width", width)
            .attr("height", this.height_content)
            .attr("x", this.scale_edges(this.scroll_spy_pos) - (width/2))
            .attr("y", "0");
          }
        }
      }
      return true;
    }
    else
      return false;
  };
  toggle_scroll_spy = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_scroll_spy(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.scroll_spy_enabled = _enable;
    else
      this.scroll_spy_enabled = !this.scroll_spy_enabled;
    return this.draw_scroll_spy();
  };

  // ‣ highlights specific segments/rectangles in yellow
  draw_search_results = function(){
    let catview = this;
    if(this.debug)
      console.log('CATview.draw_search_results');

    if(this.content){
      this.content.select(".search_results_background").selectAll('rect').remove();
      let segments = document.querySelectorAll('#'+this.parent_id+ ' .search_result');
      for (let i = 0; i < segments.length; i++) {
        segments[i].classList.remove('search_result')
      }

      if(this.search_enabled && this.search_results != null){

        if(this.search_mode === 'segment'){
          // draw highlight-rectangles over the segment-rectangles
          // first filter the hits according to the zoom
          let hits = this.search_results.filter(function(d){return (catview.from <= d[0] + 1 && d[0] +1 <= catview.to)});

          if(hits.length > 0) {
            hits.forEach(function(edge){
              edge[1].forEach(function(name_idx){
                let seg = document.querySelector('#' + catview.parent_id + ' [data-segment-index="' + name_idx + "_" + edge[0] + '"] .segment-rect')
                if(seg !== null) seg.classList.add("search_result")
              });
            });
          }
        }
        else{
          // draw full columns in the background of edges that include a search hit
          for (let i = 0; i < this.search_results.length; i++){
            let edge = this.search_results[i][0] + 1;
            // test whether the search hit is within the currently shown excerpt of the alignment
            if(this.from <= edge && edge <= this.to){
              if(this.vertical === true){
                let height = this.rect_height + 2;
                this.content.select(".search_results_background").append('rect')
                .attr('class', 'search_result')
                .attr("width", this.width_content - 4)
                .attr("height", height)
                .attr("x", 2)
                .attr("y", this.scale_edges(edge) - (height/2));

              }
              else{
                let width = this.rect_width + 2;
                this.content.select(".search_results_background").append('rect')
                .attr('class', 'search_result')
                .attr("width", width)
                .attr("height", this.height_content - 4)
                .attr("x", this.scale_edges(edge) - (width/2))
                .attr("y", 2);
              }
            }
          }
        }
      }
      return true;
    }
    else
      return false;
  };
  set_search_results = function(_search_results){
    if(_search_results == null)
      return false;

    this.search_results = _search_results;
    return true;
  };
  toggle_search = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_search(' + _enable + ', ' + _mode + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.search_enabled = _enable;
    else
      this.search_enabled = !this.search_enabled;
    return this.draw_search_results();
  };
  toggle_search_mode = function(_mode){
    if(this.debug)
      console.log('CATview.toggle_search_mode(' + _mode + ')');
    if(_mode != null && typeof(_mode) === "string"){
      if(_mode === 'segment' || _mode === 'seg')
        this.search_mode = 'segment';
      else if (_mode === 'edge' || _mode === 'col')
        this.search_mode = 'edge';
    }
    else{
      if(this.search_mode === 'segment')
        this.search_mode = 'edge'
      else // edge
        this.search_mode = 'segment'
    }
    return this.draw_search_results();
  };
  // draws the search results by highlighting rectangles with yellow bars

  // ‣ draws extra segments in a dotted manner
  // TODO: Format für extra_segments noch seg_wit statt wit_seg
  draw_extra_segments = function(){
    let catview = this;
    if(this.debug)
      console.log('CATview.draw_extra_segments');

    if(this.content) {
      this.content.select(".alignment").selectAll("rect.extra").remove();
      if (this.extra_segments == null)
        return false;

      if (this.extra_segments_enabled) {
        let rows = this.content.select(".alignment").selectAll("g.row_witness");
        this.extra_segments.forEach(function (segment) {
          let row = rows.filter(function (d, i) { return (i + catview.from) === segment[0] + 1 });

          let seg = row.append('g')
          .attr("class", "segment")
          .attr("data-segment-index",segment[1] + "_" + segment[0])
          .attr("transform", function(d) {
            if(catview.vertical === true)
              return "translate(" + (catview.scaled_position(segment[0], segment[1]) + 1) + ", 0)";
            else
              return "translate(0, " + (catview.scaled_position(segment[0], segment[1]) + 1) + ")";})
          .on("mouseenter", function(d){ catview.show_rect_linking(segment[1]+"_"+segment[0]); })
          .on("mouseleave", function(d){ catview.hide_rect_linking(); })
          .attr('cursor', 'pointer')

          seg.append("rect")
          .attr("class", "segment-rect")
          .attr("width", function(d){ return catview.scaled_rect_width(segment[0], segment[1]) - 1; })
          .attr("height", function(d){ return catview.scaled_rect_height(segment[0], segment[1]) - 2; })
          //.attr("ry", rect_corner)      // rounded Corners
          //.attr(catview.vertical === true ? "x" : "y", function(d) { return catview.scaled_position(segment[0], segment[1]) + 1; })
          .attr("style", function(d) {
            return 'fill:transparent; stroke-width:2; stroke-dasharray:3,1; stroke:' +
              (parseFloat(segment[2]) === 0.0 && catview.equality_color_enabled ? catview.equality_color : catview.scale_color(parseFloat(segment[2])));
          })
        });
      }
    }
    else
      return false;
  };
  set_extra_segments = function(_segments){
    if(_segments == null)
      return false;

    // todo validate data
    this.extra_segments = _segments;
    return true;
  };
  toggle_extra_segments = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_extra_segments(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.extra_segments_enabled = _enable;
    else
      this.extra_segments_enabled = !this.extra_segments_enabled;
    return this.draw_extra_segments();
  };

  // ‣ draws the icons for tools included in CATview.tools
  draw_tools = function(){
    let catview = this;
    if(this.debug)
      console.log('CATview.draw_tools');

    // remove current tool icons
    let toolicons = this.svg.select(".tool-icons");
    toolicons.selectAll('*').remove();

    if(this.tools_enabled){
      let margin = 3;     // extra margin around the tool icon container
      let width, height;  // available space for the tool icon container

      // place the container for the tool icons
      if(this.vertical === true){
        width = this.width_svg - this.margin.left - this.margin.right - 2 * margin;
        height = this.space_for_tools - 2 * margin;
        if(this.x_inverted)
          toolicons.attr("transform", "translate(" + (this.margin.left + margin) + ", " + margin + ")");
        else
          toolicons.attr("transform", "translate(" + (this.margin.left + margin) +  ", " + (this.height_svg - this.margin.bottom + margin) + ")");
      } else {
        width = this.space_for_tools - 2 * margin;
        height = this.height_svg - this.margin.top - this.margin.bottom - 2 * margin;
        if(this.x_inverted)
          toolicons.attr("transform", "translate(" + margin + ", " + (this.margin.top + margin) + ")");
        else
          toolicons.attr("transform", "translate(" + (this.width_svg - this.margin.right + margin) +  ", " + (this.margin.top + margin) + ")");
      }
      /*toolicons.append('rect')
        .attr("x", -margin)
        .attr("y", -margin)
        .attr("width", width + 2 * margin)
        .attr("height", height + 2 * margin)
        .attr('fill', '#ccccff');
      toolicons.append('rect')
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", width)
        .attr("height", height)
        .attr('fill', '#8888ff');*/

      let n = this.tools.length;  // number of tool icons to be placed
      if(n > 0) {
        let padding = 12;
        let fontsize;
        if(this.vertical){
          fontsize = (width - (n-1)*padding) / n;
          if(height < fontsize){
            fontsize = height;
            padding = (width - n * fontsize) / (n-1)
          }
        } else {
          fontsize = (height - (n-1)*padding) / n;
          if(width < fontsize){
            fontsize = width;
            padding = (height - n * fontsize) / (n-1)
          }
        }

        for(let i = 0; i < this.tools.length; i++){
          let x, y;
          if(this.vertical === true){
            x =  i * (fontsize + padding) + fontsize/2;
            y = fontsize/2;
          } else {
            x = fontsize/2;
            y =  i * (fontsize + padding) + fontsize/2;
          }

          toolicons.append('text')
          .attr("class", "tool-icon")
          .attr("x", x)
          .attr("y", y)
          .attr("style", "text-anchor: middle; dominant-baseline: middle;")
          .attr("font-family","FontAwesome")
          //.attr("textLength", fontsize)
          //.attr("lengthAdjust","spacingAndGlyphs")
          .text(String.fromCharCode(parseInt(this.tools[i][0], 16)))
          .attr('font-size', fontsize - 4)
          .attr("transform", (function(){ return 'rotate(' + catview.tools[i][2] + ' ' + x + ' ' + y + ')' }))
          .on("click", this.tools[i][1]);
        }
      }
    }
    return true
  };
  toggle_tools = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_tools(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.tools_enabled = _enable;
    else
      this.tools_enabled = !this.tools_enabled;
    return this.draw_tools();
  };

  // ‣ scaling of segments/rectangles
  set_rect_scaling = function(_data){
    if(this.debug)
      console.log('CATview.set_rect_scaling');
    // TODO validation
    this.rect_scaling_data = _data;
    return true;
  };
  toggle_rect_scaling = function(_enable, _mode){
    if(this.debug)
      console.log('CATview.toggle_rect_scaling(' + _enable + ', ' + _mode + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.rect_scaling_enabled = _enable;
    else
      this.rect_scaling_enabled = !this.rect_scaling_enabled;
    return this.refresh_content(this.from, this.to);
  };
  toggle_rect_scaling_mode = function(_mode){
    if(this.debug)
      console.log('CATview.toggle_rect_scaling_mode(' + _mode + ')');
    if(_mode != null && typeof(_mode) === "string"){
      if(_mode === 'middle' )
        this.rect_scaling_mode = 'middle';
      else if (_mode === 'inverted')
        this.rect_scaling_mode = 'inverted';
      else if (_mode === 'default')
        this.rect_scaling_mode = 'default';
    }
    else{
      if(this.rect_scaling_mode === 'default')
        this.rect_scaling_mode = 'middle'
      else if(this.rect_scaling_mode === 'middle')
        this.rect_scaling_mode = 'inverted'
      else // inverted
        this.rect_scaling_mode = 'default'
    }
    return this.refresh_content(this.from, this.to);
  }
  // internal methods to define the dimensions and positioning of rectabgles in respect to rectangle scaling
  scaled_rect_width = function(_egde_idx, _name_idx){
    if(!this.rect_scaling_enabled || !this.vertical || this.rect_scaling_data === null){
      // return the default width on disabled scaling or horizontal orientation
      return this.rect_width;
    }

    let scaling = this.rect_scaling_data[_egde_idx][_name_idx];  // individual scaling of the segment/rectangle

    let minimal_size = this.rect_width * this.rect_scaling_minimum; // fixed minimal size of the rectangle
    let addable_size = this.rect_width - minimal_size;                 // variable size available to add
    return Math.round(minimal_size + scaling * addable_size);
  };
  scaled_rect_height = function(_egde_idx, _name_idx){
    if(!this.rect_scaling_enabled || this.vertical || this.rect_scaling_data === null){
      // return the default height on disabled scaling or vertical orientation
      return this.rect_height;
    }

    let scaling = this.rect_scaling_data[_egde_idx][_name_idx];  // individual scaling of the segment/rectangle

    let minimal_size = this.rect_height * this.rect_scaling_minimum; // fixed minimal size of the rectangle
    let addable_size = this.rect_height - minimal_size;                 // variable size available to add
    return Math.round(minimal_size + scaling * addable_size);
  };
  scaled_position = function(_egde_idx, _name_idx){
    if(!this.rect_scaling_enabled || this.rect_scaling_mode == 'inverted'){
      // the default value if scaling was disabled or on inverted mode
      return this.scale_names(this.drag_order_name2pos[_name_idx]);
    }

    if(this.vertical){
      let scaled = this.scaled_rect_width(_egde_idx, _name_idx)
      if(this.rect_scaling_mode === 'middle')
        scaled = (scaled + this.rect_width) / 2.0;
      // the original posittion + maximal size - used size
      return this.scale_names(this.drag_order_name2pos[_name_idx]) + this.rect_width - scaled
    }else{
      // get the scaled size
      let scaled = this.scaled_rect_height(_egde_idx, _name_idx)
      if(this.rect_scaling_mode === 'middle')
        scaled = (scaled + this.rect_height) / 2.0;
      // the original posittion + maximal size - used size
      return this.scale_names(this.drag_order_name2pos[_name_idx]) + this.rect_height - scaled
    }
  };


  // • interactive elements
  // ‣ react on dragging a name
  // _name2pos = array that maps: name index → position index, e.g. [3,1,2,0] (swap first and last text)
  set_names_order = function(_name2pos){
    if(this.debug)
      console.log('CATview.set_names_order');

    // TODO validate data
    this.drag_order_name2pos = _name2pos;

    if(this.drag_enabled){
      this.draw_names_axis();
      this.refresh_content(this.from, this.to);
    }

    return true;
  };
  reset_names_order = function(){
    if(this.debug)
      console.log('CATview.reset_names_order');

    let names_order = {};
    for(let i = 0; i < this.names.length; i++)
      names_order[i] = i;

    return this.set_names_order(names_order);
  };
  // returns an array that maps: name index → position index, e.g. [3,1,2,0] (swap first and last text)
  get_names_order = function(){
    if(this.debug)
      console.log('CATview.get_names_order');

    return this.drag_order_name2pos;
  };
  set_drag_callback = function(_callback) {
    if(this.debug)
      console.log('CATview.set_drag_callback');

    this.drag_callback = _callback;
    return true;
  };
  toggle_drag = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_drag(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.drag_enabled = _enable;
    else
      this.drag_enabled = !this.drag_enabled;
    // add/remove the drag behaviour to/from the axis
    if(this.svg !== null && this.svg.select(".names-axis").size() === 1 ){
      let names_axis = document.querySelector('#' + this.parent_id + ' .names-axis');
      if(this.drag_enabled)
        names_axis.classList.add('draggable');
      else
        names_axis.classList.remove('draggable');
      return this.draw_names_axis();
    }
    return false;
  };
  toggle_drag_mode = function(_mode){
    if(this.debug)
      console.log('CATview.toggle_drag_mode(' + _mode + ')');
    if(_mode != null && typeof(_mode) === "string"){
      if(_mode === 'insert' )
        this.drag_mode = 'insert';
      else if (_mode === 'swap')
        this.drag_mode = 'swap';
    }
    else{
      if(this.drag_mode === 'insert')
        this.drag_mode = 'swap'
      else
        this.drag_mode = 'insert'
    }
    return this.draw_names_axis();
  };

  // ‣ display an edges name when hovering over its rectangles
  set_edge_names = function(_names){
    if(this.debug)
      console.log('CATview.set_edge_names');
    if(_names == null)
      return false;
    this.edge_names = _names;
    return true;
  };
  toggle_edge_names = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_edge_names(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.edge_names_enabled = _enable;
    else
      this.edge_names_enabled = !this.edge_names_enabled;
    return true;
  };
  show_edge_name = function (_pos) {
    if(this.debug)
      console.log('CATview.show_edge_name: ' + _pos);

    if(!this.edge_names_enabled)
      return false;

    // respect current zooming
    _pos += this.from - 1;

    this.edge_names_box.transition()
      .duration(200)
      .style("opacity", 1.0);

    if(this.edge_names[_pos] !== null && this.edge_names[_pos] !== undefined && this.edge_names[_pos] !== '')
      name = (_pos + 1).toString() + ' • ' + this.edge_names[_pos];
    else
      name = 'Position: ' + (_pos + 1).toString();

    // place the edge name central in respect to CATviews orientation
    let anchor_x = "left";
    let offset_x = "50%";
    let anchor_y = "top";
    let offset_y = document.getElementById(this.parent_id).offsetTop + this.height_svg + 'px';
    let transform = "translate(-50%, 0)";
    switch(this.orientation) {
      case 'left':
        offset_x = this.width_svg + 'px';
        offset_y = "50%";
        transform = "translate(0, -50%)";
        break;
      case 'bottom':
        anchor_y = "bottom";
        // CATviews height + the distance to the bottom of the window
        offset_y = this.height_svg + window.innerHeight - document.getElementById(this.parent_id).getBoundingClientRect().bottom + 'px';
        break;
      case 'right':
        anchor_x = "right";
        offset_x = this.width_svg + 'px';
        offset_y = "50%";
        transform = "translate(0, -50%)";
        break;
      default: // top
    }
    this.edge_names_box.html(name)
      .style(anchor_x, offset_x)
      .style(anchor_y, offset_y)
      .style("transform", transform);
  };
  hide_edge_name = function () {
    if(this.debug)
      console.log('CATview.hide_edge_name');
    this.edge_names_box.transition()
      .duration(200)
      .style("opacity", 0);
  };

  // ‣ linking of segments/rectangles which will highligted on hovering
  set_rect_linking_data = function(_data){
    if(this.debug)
      console.log('CATview.set_rect_linking_data');
    // TODO validation
    this.rect_linking_data = _data;
    return true;
  };
  toggle_rect_linking = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_rect_linking(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.rect_linking_enabled = _enable;
    else
      this.rect_linking_enabled = !this.rect_linking_enabled;

    if(this.rect_linking_enabled)
      return true
    else
      return this.hide_rect_linking();
  };
  // internal methods for rectangle linking
  show_rect_linking = function(_idx) {
    let catview = this;
    if(this.debug)
      console.log('CATview.show_rect_linking: ' + _idx);

    if(!this.rect_linking_enabled)
      return false;

    this.rect_linking_data.forEach( function (group){
      if(group.indexOf(_idx) !== -1){
        group.forEach( function (group_member) {
          // add the 'rect_linked' class to highlight all segments in the group, including the one hovered
          let seg = document.querySelector('#' + catview.parent_id + ' [data-segment-index="' + group_member + '"] .segment-rect');
          if(seg !== null) seg.classList.add("rect_linked");
        });
      }
    });
    
    // optinal callback 
    if (this.show_rect_linking_callback !== null)
      this.show_rect_linking_callback(this, _idx);

  };
  hide_rect_linking = function(){
    if(this.debug)
      console.log('CATview.hide_rect_linking');

    // remove all 'rect_linked' classes
    let segments = document.querySelectorAll('#'+this.parent_id + ' .rect_linked');
    for (let i = 0; i < segments.length; i++) {
      segments[i].classList.remove('rect_linked');
    }
    
    // optional callback
    if (this.hide_rect_linking_callback !== null)
      this.hide_rect_linking_callback(this);

  };
  // setters for callbacks
  set_show_rect_linking_callback = function(_callback) {
    this.show_rect_linking_callback = _callback;
  }
  set_hide_rect_linking_callback = function(_callback) {
    this.hide_rect_linking_callback = _callback;
  }

  // ‣ zooming within the content area of CATview
  // set the maximum zooming level
  set_max_zoom = function(_max){
    let max = parseFloat(Number(_max));
    if(isNaN(max)){
      // the default value will allow to increase the size of edges as far as only ten of them can be displayed
      max = this.edges.length / 10;
    }
    else if (max < 1 ) {
      // setting the maximum scale to 1 will disable the zoom
      max = 1.0;
    }

    this.zoom.scaleExtent([1, max]);

    // todo react if current zoom bigger than max?

    return true;
  };
  toggle_zoom = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_zoom(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.zoom_enabled = _enable;
    else
      this.zoom_enabled = !this.zoom_enabled;

    if(this.zoom_enabled)
      this.content.call(this.zoom);
    else
      this.content.on(".zoom", null);

     return true;
  }
  zoom_to = function(_factor) {
    this.zoom_button_clicked = true;
    if (this.zoom.scaleExtent()[0] <= _factor && _factor <= this.zoom.scaleExtent()[1]){
      this.zoom.scaleTo(d3.select(this.content).node(), _factor);
    }
  };
  zoom_in = function() {    
    this.zoom_button_clicked = true;
    if (this.scale+this.zoom_step <= this.zoom.scaleExtent()[1]){
      this.zoom.scaleTo(d3.select(this.content).node(), this.scale+this.zoom_step);   
    }
    else {
      // we hit a boundary, might as well set scale to the boundary
      // sometimes mixed scaling with mouse wheel and buttons causes the zoom-step to be
      // too large so it would exceed the boundary. In that case clip to the boundary
      this.zoom.scaleTo(d3.select(this.content).node(), this.zoom.scaleExtent()[1]);
    }    
  };
  zoom_out = function() {
    this.zoom_button_clicked = true;
    if (this.scale-this.zoom_step >= this.zoom.scaleExtent()[0])
      this.zoom.scaleTo(d3.select(this.content).node(), this.scale-this.zoom_step);
    else
      this.zoom.scaleTo(d3.select(this.content).node(), this.zoom.scaleExtent()[0]);   
  };
  zoom_reset = function () {
    this.zoom_button_clicked = true;
    d3.select(this.content).node().call(this.zoom.transform, d3.zoomIdentity);
  }
  // callback with functionality for scaling and translation of the content by mousewheel
  zooming = function () {
    if(this.debug)
      console.log('CATview.zooming');

    let new_scale = d3.event.transform.k;    
    let new_translate = this.vertical === true ? (d3.event.transform.y/new_scale) : (d3.event.transform.x/new_scale);

    if(this.content){
      let delta_translate = 0;
      let scale_from = 0;
      let scale_to = 0;
      let current_size = this.vertical === true ? this.height_content : this.width_content;
      // if the scale factor has changed
      if (new_scale !== this.scale) {
        // calculate the change of the size
        let delta_size = (current_size / new_scale) - (current_size / this.scale);
        // distribute the change of the size among the left/top and right/bottom site due to the relative mouse center

        // simulate central mouse position in case of zoom button clicks
        if (this.zoom_button_clicked){
          let art_mouse_x = this.width_content/2;
          let art_mouse_y = this.height_content/2;
          scale_from = delta_size * ((this.vertical === true ? art_mouse_y : art_mouse_x) / current_size);
          this.zoom_button_clicked = false;
        }
        else{
          // get the content container (the zoom event is attached here)
          let content = this.content._groups[0][0];
          scale_from = delta_size * ((this.vertical === true ? d3.mouse(content)[1] : d3.mouse(content)[0]) / current_size);
        }
        scale_to = delta_size - scale_from;
      }
      else if(this.brush_enabled === false){
        // calculate the relative translation (only if the brush function is disabled)
        delta_translate = this.translate - new_translate;
      }
      
      // calculate the new left and right position in the original scale (0 upto size)
      let offset = 0;     // if the limit of one site is reached, add the remaining pixels to the other site
      let new_from_pixel = this.from_pixel - scale_from + delta_translate;
      if (new_from_pixel < 0)       // the limit of the left site was reached
      {
        offset = -new_from_pixel;
        new_from_pixel = 0;
      }
      let new_to_pixel = this.to_pixel + scale_to + delta_translate + offset;
      if (new_to_pixel > current_size)  // the limit of the right site was reached
      {
        offset =  new_to_pixel - current_size;
        new_to_pixel = current_size;
        new_from_pixel -= offset;
      }

      // return when one of the new interval borders is not legit (e.g. by double clicking a rectangle)
      if(isNaN(new_from_pixel) || isNaN(new_to_pixel))
        return false;

      // save the parameters of this call
      this.from_pixel = new_from_pixel;
      this.to_pixel = new_to_pixel;
      this.translate = new_translate;
      this.scale = new_scale;

      // calculate the edges within this new zoom
      let new_from = Math.ceil(this.scale_edges_original.invert(new_from_pixel));
      let new_to = Math.floor(this.scale_edges_original.invert(new_to_pixel));
      if(this.x_inverted){
        let swap = new_from;
        new_from = new_to;
        new_to = swap;
      }
      if(new_from !== this.from || new_to !== this.to) {
        // refresh the content only if edges within the new zoom have changed
        this.refresh_content(new_from, new_to);
      }

      // refresh the brush
      if(this.brush) {
        let from = this.brush_from_edge + 1.0 + this.brush_offset_from;
        let to = this.brush_to_edge + 1.0 + this.brush_offset_to;
        if(this.x_inverted){
          let swap = from;
          from = to;
          to = swap;
        }
        this.brush.move(d3.select('.brush'), [from, to].map(this.scale_edges))
        // TODO prevent brushing listeners
      }
    }
    else
      return false;
  };
  // fade out borders when zooming in
  toggle_fadeout = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_fadeout(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.fadeout_enabled = _enable;
    else
      this.fadeout_enabled = !this.fadeout_enabled;
    return this.draw_fadeout();
  };
  draw_fadeout = function(){
    if(this.debug)
      console.log('CATview.draw_fadeout');

    // size of gradient compared to content size
    let grad_fraction = 0.5;

    d3.select("#"+this.parent_id+"_fadeout_left").remove();
    d3.select("#"+this.parent_id+"_fadeout_right").remove();

    if(this.fadeout_enabled) {
      // fadeout left / top
      //if (this.from > 1) {
        // je kleiner der Anteil von 0 bis from, desto kleiner soll auch
        // der fade out in der Breite sein.
        let zoom_from_fraction = ((this.from-1)/this.edges.length);
        if(this.x_inverted)
          zoom_from_fraction = ((this.edges.length-this.to)/this.edges.length);

        this.content.select(".alignment")
        .append("rect")
        .attr("id", this.parent_id+"_fadeout_left")
        .attr("height", this.vertical === true ? Math.floor((this.height_content*grad_fraction)*zoom_from_fraction) : this.height_content-2 )
        .attr("width", this.vertical === true ? this.width_content-2 : Math.floor((this.width_content*grad_fraction)*zoom_from_fraction))
        .attr("transform", "translate" +
          (this.vertical === true ? "(1,1)" : "(1,1)"))
        .style("fill", "url(#"+this.parent_id+"_gradient_left)")
        .style("pointer-events", "none");
      //}

      // fadeout right / bottom
      //if (this.to < this.edges.length) {
        let zoom_to_fraction = ((this.edges.length-this.to)/this.edges.length);
        if(this.x_inverted)
          zoom_to_fraction = ((this.from-1)/this.edges.length);

        this.content.select(".alignment")
        .append("rect")
        .attr("id", this.parent_id+"_fadeout_right")
        .attr("height", this.vertical === true ? Math.floor(this.height_content*grad_fraction*zoom_to_fraction) : this.height_content-2 )
        .attr("width", this.vertical === true ? this.width_content-2 : Math.floor(this.width_content*grad_fraction*zoom_to_fraction))
        .attr("transform", "translate("+
          (this.vertical === true ? ("1, " + Math.floor(this.height_content-this.height_content*grad_fraction*zoom_to_fraction-1)) :
              (Math.ceil(this.width_content-this.width_content*grad_fraction*zoom_to_fraction-1) +", 1")
          )+")")
        .style("fill", "url(#"+this.parent_id+"_gradient_right)")
        .style("pointer-events", "none");

      //}
    }
    return true;
  }

  // ‣ react on a click on a rectangle
  toggle_click_on_edge = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_click_on_edge(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.click_on_edge_enabled = _enable;
    else
      this.click_on_edge_enabled = !this.click_on_edge_enabled;
    // refresh the cursors for rectangles
    let segments = document.querySelectorAll('#'+this.parent_id + ' .segment-rect');
    if(this.click_on_edge_enabled)
      for (let i = 0; i < segments.length; i++)
        segments[i].setAttribute('cursor', 'pointer');
    else
      for (let i = 0; i < segments.length; i++)
        segments[i].setAttribute('cursor', 'inherent');

    return true
  };
  set_click_on_edge_callback = function(_callback) {
    this.click_on_edge_callback = _callback;

    return true;
  };

  // ‣ the brushing tool for selecting specific segments via a box
  draw_brush = function(){
    if(this.debug)
      console.log('CATview.draw_brush');

    if(this.content){
      // remove previous brush
      this.content.select(".brush").remove();

      if(this.brush_enabled){
        // add the brush to CATview

        if(this.vertical === true) {
          this.brush = d3.brushY()
          .on("brush", this.brushing.bind(this))
          .on("end", this.brush_end.bind(this));
          // bind the CATview object to make it directly available within brushing() and brush_end()

          if(this.x_inverted)
            this.brush.extent([[this.scale_names.range()[0], this.scale_edges.range()[1]], [this.scale_names.range()[1], this.scale_edges.range()[0]]]);
          else
            this.brush.extent([[this.scale_names.range()[0], this.scale_edges.range()[0]], [this.scale_names.range()[1], this.scale_edges.range()[1]]]);
        } else {
          this.brush = d3.brushX()
          .on("brush", this.brushing.bind(this))
          .on("end", this.brush_end.bind(this));
          // bind the CATview object to make it directly available within brushing() and brush_end()

          if(this.x_inverted)
            this.brush.extent([[this.scale_edges.range()[1], this.scale_names.range()[0]], [this.scale_edges.range()[0], this.scale_names.range()[1]]]);
          else
            this.brush.extent([[this.scale_edges.range()[0], this.scale_names.range()[0]], [this.scale_edges.range()[1], this.scale_names.range()[1]]]);
        }

        this.content.append("g")
        .attr("class", "brush")
        .call(this.brush);
      }
      return true;
    }
    else
      return false;
  };
  // place brush at a specific position (defined by indices)
  place_brush = function(_from, _to, _trigger_callback, _trigger_end_callback){
    if(this.debug)
      console.log('CATview.place_brush: ' + _from + ', ' + _to);

    // enable brush as it might be currently uninitialized
    this.toggle_brush(true);

    // save the new interval
    let brush_range = _to - _from + 1;
    this.brush_from_edge = _from;
    this.brush_to_edge = _to;
    this.brush_range = brush_range;

    // draw the new interval
    if(this.x_inverted) {
      this.brush.move(d3.select(".brush"),
        [_to + 1.0 + this.brush_offset_to, _from + 1.0 + this.brush_offset_from].map(this.scale_edges));
    }
    else {
      this.brush.move(d3.select(".brush"),
        [_from + 1.0 + this.brush_offset_from, _to + 1.0 + this.brush_offset_to].map(this.scale_edges))
    }

    // may trigger callbacks
    if (_trigger_callback && this.brush_callback !== null)
      this.brush_callback(this, _from, _to); //, from_name, to_name);
    if (_trigger_end_callback && this.brush_end_callback !== null)
      this.brush_end_callback(this, _from, _to); //, from_name, to_name);

    return true;
  }
  toggle_brush = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_brush(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.brush_enabled = _enable;
    else
      this.brush_enabled = !this.brush_enabled;

    // pass the full interval to the callback on remove
    if(!this.brush_enabled && this.brush_callback != null && this.edges.length > 0)
      this.brush_callback(this, 0, this.edges.length - 1, 0, this.names.length - 1);

    return this.draw_brush();
  };
  // set the callback after brushing
  set_brush_callback = function(_callback) {
    this.brush_callback = _callback;

    return true;
  };
  set_brush_end_callback = function(_callback) {
    this.brush_end_callback = _callback;

    return true;
  };
  // react on changes of the brush
  brushing = function() {
    if(this.debug)
      console.log('CATview.brushing');

    // handle inverted edges axis
    let from = this.x_inverted ? 1 : 0;

    // the starting position is construction by:
    //    d3.event.selection[0] → current begin of the drawn selection window
    //    this.brush_offset_from → the specified offset used to place the window at the begin of a rectangle
    //    0.001 → a little additional offset for the case that the window was dragged to the most right position
    let from_edge = Math.round(this.scale_edges.invert(d3.event.selection[from]) - this.brush_offset_from - 0.001) -1;
    if(from_edge < this.from - 1) from_edge = this.from - 1;
    else if(from_edge >= this.to) from_edge = this.to - 1;

    // add the current range to the from position to get the to position
    let to_edge = from_edge + this.brush_range - 1;

    // TODO may allow selection of names axis → brush_mode: edges | names | both
    //let from_name = null; // Math.ceil(this.brush.extent()[0][names_index]);
    //let to_name = null; // Math.floor(this.brush.extent()[1][names_index]);

    if(to_edge - from_edge >= 0) {
      // check whether the interval has changed
      if (from_edge !== this.brush_from_edge || to_edge !== this.brush_to_edge){ // || from_name != this.brush_from_name || to_name != this.brush_to_name) {
        // save the current interval
        this.brush_from_edge = from_edge;
        this.brush_to_edge = to_edge;
        //this.brush_from_name = from_name;
        //this.brush_to_name = to_name;
        // call the brush_callback as the interval has changed
        if (this.brush_callback !== null)
          this.brush_callback(this, from_edge, to_edge, null, null); //, from_name, to_name);
        else {
          // no callback specified → log positions to console
          console.log('CATview brush edge from ' + from_edge + ' to ' + to_edge);// + ' and name from ' + from_name + ' to ' + to_name);
        }
      }
    }
  };
  // react on the end of a brush gesture
  brush_end = function() {
    if(this.debug)
      console.log('CATview.brush_end');

    if (!d3.event.sourceEvent) return; // only transition after input
    if (!d3.event.selection) return; // ignore empty selections

    // handle inverted edges axis
    let from = this.x_inverted ? 1 : 0;
    let to = this.x_inverted ? 0 : 1;

    // the starting position (end analogously) is construction by:
    //    d3.event.selection[0] → current begin of the drawn selection window
    //    this.brush_offset_from → the specified offset used to place the window at the begin of a rectangle
    //    0.001 → a little additional offset for the case that the window was dragged to the most right position
    let from_edge = Math.round(this.scale_edges.invert(d3.event.selection[from]) - this.brush_offset_from - 0.001) -1;
    if(from_edge < this.from - 1) from_edge = this.from - 1;
    else if(from_edge >= this.to) from_edge = this.to - 1;
    let to_edge = Math.round(this.scale_edges.invert(d3.event.selection[to]) - this.brush_offset_to + 0.001) -1;
    if(to_edge >= this.to) to_edge = this.to - 1;
    else if(to_edge < this.from - 1) to_edge = this.from - 1;

    // calculated the number of rectangles within the window
    let brush_range = to_edge - from_edge + 1;

    // intervene if range is zero
    if(brush_range === 0){
      to_edge = from_edge;
      brush_range = 1;
    }

    // smoothly redraw the brush to discrete positions (begin and end of the rectangles)   
    if(this.x_inverted){
      d3.select('#' + this.parent_id + ' .brush ').transition().call(d3.event.target.move,
        [to_edge + 1.0 + this.brush_offset_to, from_edge + 1.0 + this.brush_offset_from].map(this.scale_edges));
    }else{
      d3.select('#' + this.parent_id + ' .brush ').transition().call(d3.event.target.move,
        [from_edge + 1.0 + this.brush_offset_from, to_edge + 1.0 + this.brush_offset_to].map(this.scale_edges));
    }

    // save new values if there was a change
    if(from_edge !== this.brush_from_edge || to_edge !== this.brush_to_edge || brush_range !== this.brush_range){
      // save the current interval
      this.brush_from_edge = from_edge;
      this.brush_to_edge = to_edge;
      this.brush_range = brush_range;
      // and call the custom callbacks
      if (this.brush_end_callback != null)
        this.brush_end_callback(this, from_edge, to_edge); //, from_name, to_name);
    }
  };

  // ‣ custom tools that will be triggered when clicking their icon
  // add a new tool icon and its behavior
  add_tool = function(_icon, _callback, _index, _rotation){
    let index = parseInt(_index);
    if (isNaN(index) || index < 0 || index > this.tools.length )
      index = this.tools.length;

    let rotation = (typeof _rotation === 'undefined') ? 0 : parseInt(_rotation);

    this.tools.splice(index, 0, [_icon, _callback, rotation]);

    return true;
  };
  // remove a tool
  remove_tool = function(_index){
    let index = parseInt(Number(_index));
    if (isNaN(index) || index < 0 || index >= this.tools.length )
      return false;

    this.tools.splice(index, 1);

    return true;
  };
  // add the invert-x-axis tool
  add_tool_invert_edges_axis = function(_index){
    let catview = this;
    return this.add_tool('f0dc',
      function(){catview.invert_edges_axis();},
      _index,
      catview.vertical ? 0 : 90);
  };
  // add the invert-y-axis tool
  add_tool_invert_names_axis = function(_index){
    let catview = this;
    return this.add_tool('f0dc',
      function(){catview.invert_names_axis();},
      _index,
      catview.vertical ? 90 : 0);
  };
  // add the switch-highlight-mode-of-search-results tool
  add_tool_toggle_search_mode = function(_index){
    let catview = this;
    return this.add_tool('f002',
      function(){catview.toggle_search_mode()},
      _index,
      0);
  };
  // add the enable/disable-brush tool
  add_tool_toggle_brush = function(_index){
    let catview = this;
    return this.add_tool('f1fc', // f096
      function(){catview.toggle_brush();},
      _index,
      0);
  };
  // add a tool that toggles the rectangle scaling
  add_tool_toggle_rect_scaling = function(_index){
    let catview = this;
    return this.add_tool(
      'f037', // (this.vertical ? 'f036' : 'f038'),
      function(){catview.toggle_rect_scaling()},
      _index,
      (catview.vertical ? 0 : 90 ));
  };
  // zoom in and out buttons
  add_tool_zoom_in = function(_index){
    let catview = this;
    return this.add_tool('f00e', // f067
      function(){catview.zoom_in();},
      _index,
      0);
  };
  add_tool_zoom_out = function(_index){
    let catview = this;
    return this.add_tool('f010', // f068
      function(){catview.zoom_out();},
      _index,
      0);
  };


  // • colors and appearance
  set_color_scale = function(_from, _to, _border){

    if(_from === null || _from === undefined) _from = "#c1c1e9";
    if(_to === null || _to === undefined) _to = "#0000a3";
    if(_border === null || _border === undefined) _border = "#2f2f86";

    this.scale_color = d3.scaleLinear().domain([0, 1]).range([_from, _to]);

    this.rect_border = "stroke: " + _border + ";";

    return this.refresh_content();
  };
  set_background_color = function(_color){
    if(_color == null)
      return false;

    this.background = _color;
    document.querySelector('#' + this.parent_id + ' svg').style.backgroundColor = _color;
    document.querySelector('#' + this.parent_id + ' .edge-name').style.backgroundColor = _color;
    // TODO also update the fade outs
    return true;
  };
  toggle_equality_color = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_equality_color(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.equality_color_enabled = _enable;
    else
      this.equality_color_enabled = !this.equality_color_enabled;
    return this.draw_alignment();
  };


  // • enable/disable/toggle printing debug information on console
  toggle_debug = function(_enable){
    console.log('CATview.toggle_debug(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.debug = _enable;
    else
      this.debug = !this.debug;
    return true;
  };

  // alignment
  //   g-row        → events, position
  //     g-segment  → identifier, events, position
  //       rect       → classes: extra, search
  //       text
  //   gradienten
  // TODO: handle font size and placement in vertical orientation
  draw_segment_texts = function(){
    if(this.debug)
      console.log('CATview.draw_segment_texts(' + _enable + ')');

    let approved = true
    if(this.segment_texts_enabled) {
      if(this.segment_texts_min_scale === null || this.segment_texts_min_scale <= this.scale ) {
        if(this.segment_texts_min_width !== null && this.segment_texts_min_width > this.rect_width)
          approved = false;
      } else approved = false;
    } else approved = false;
    //console.log(this.segment_texts_enabled + ' ' + this.scale + ' ' + this.rect_width + ' → ' + approved);

    if(approved){
      let font_size = this.rect_height / 2;

      // placement depending of the writing direction
      let anchor = 'start';
      let x =  2;  //  + 2 px extra margin
      if(this.segment_texts_mode == 'r2l'){
        anchor = 'end'
        x = this.rect_width - 2;
      }

      for(let i = this.from; i <= this.to; i++ ) {
        // simple heuristic based on the monospace font; assuming each letter has a width of 2/3 of font size
        let num_letters = Math.floor(3 * this.rect_width / font_size / 2);

        for (let j = 0; j < this.edges[i - 1].length; j++) {
          if (this.edges[i - 1][j] != -1) {
            let y = this.rect_height / 2 + font_size / 3; // centered
            //let text = this.edge_names[i - 1] ? this.edge_names[i - 1].substr(0, num_letters) : '‧';
            let text = this.segment_texts[i-1][j] ? this.segment_texts[i-1][j].substr(0, num_letters) : ''
            // text = 'וכן יעקב אמת. שנקרא כסא'.substr(0, num_letters); // rtl

            d3.select('#' + this.parent_id + ' [data-segment-index="' + j + "_" + (i-1) + '"] text')
            .text(text)
            .attr("x", x)
            .attr("y", y)
            .attr('font-size', font_size)
            .attr("style", "text-anchor: " + anchor + ";")
          }
        }
      }
    }

    return true;
  }

  set_segment_texts = function(_data){
    if(_data == null)
      return false;

    // todo validate data
    this.segment_texts = _data;
    return true;
  };
  toggle_segment_texts = function(_enable){
    if(this.debug)
      console.log('CATview.toggle_segment_texts(' + _enable + ')');
    if(_enable != null && typeof(_enable) === "boolean")
      this.segment_texts_enabled = _enable;
    else
      this.segment_texts_enabled = !this.segment_texts_enabled;
    // reset all texts within segments
    if(!this.segment_texts_enabled)
      this.content.selectAll(".segment-text").text('')
    return this.draw_segment_texts();
  };
  toggle_segment_texts_mode = function(_mode){
    if(this.debug)
      console.log('CATview.toggle_segment_texts_mode(' + _mode + ')');
    if(_mode != null && typeof(_mode) === "string"){
      if(_mode === 'l2r' )
        this.segment_texts_mode = 'l2r';
      else if (_mode === 'r2l')
        this.segment_texts_mode = 'r2l';
    }
    else{
      if(this.segment_texts_mode == 'l2r')
        this.segment_texts_mode = 'r2l';
      else
        this.segment_texts_mode = 'l2r';
    }
    return this.draw_segment_texts();
  };


  // todo
  // - may: method for changing the border_color and others → set_border(_color, _width)
  // - adopt font size in respect to the number of witnesses
  // - method to reset all data → reset

  // TODO evtl? + reset_brush → analog zu disable_brush
  reset_brush = function(_trigger_callback, _trigger_end_callback){
    this.brush_from_edge = -1;
    this.brush_to_edge = -1;
    this.brush_from_name = -1;
    this.brush_to_name = -1;
    this.brush_range = 1; // ?
    // todo: full range oder ganz weglassen?
    // may trigger callbacks
    //if (_trigger_callback && this.brush_callback !== null)
    //  this.brush_callback(this, _from, _to); //, from_name, to_name);
    //if (_trigger_end_callback && this.brush_end_callback !== null)
    //  this.brush_end_callback(this, _from, _to); //, from_name, to_name);
    return true;
  }

  // deprecated
  enable_rect_scaling = function(_mode){
    console.log('CATview.enable_rect_scaling IS DEPRECAETED');

    this.rect_scaling_enabled = true;
    if(_mode === 'middle' )
      this.rect_scaling_mode = 'middle';
    else if (_mode === 'inverted')
      this.rect_scaling_mode = 'inverted';
    else
      this.rect_scaling_mode = 'default';
    this.refresh_content(this.from, this.to);
    return true;
  };
  disable_rect_scaling = function(){
    console.log('CATview.disable_rect_scaling IS DEPRECAETED');

    this.rect_scaling_enabled = false;
    this.refresh_content(this.from, this.to);
    return true;
  };
  enable_rect_linking = function(){
    if(this.debug)
      console.log('CATview.enable_rect_linking');

    this.rect_linking_enabled = true;

    return true;
  };
  disable_rect_linking = function(){
    if(this.debug)
      console.log('CATview.disable_rect_linking');

    this.rect_linking_enabled = false;
    this.hide_rect_linking();
    return true;
  };
  // enable the brush functionality and pass
  // a callback function to the brush (will receive: from_edge, to_edge, from_name, to_name)
  enable_brush = function(_brush_callback, _brush_end_callback){
    if(this.debug)
      console.log('CATview.enable_brush');

    if(this.brush_enabled === false){
      this.brush_enabled = true;
      // add the brush to CATview
      if(this.content)
        this.draw_brush();
    }
    if(_brush_callback !== null)
      this.brush_callback = _brush_callback;
    if(_brush_end_callback !== null)
      this.brush_end_callback = _brush_end_callback;

    return true;
  };
  // disable the brush functionality and set the attributes back in default state
  disable_brush = function(){
    if(this.debug)
      console.log('CATview.disable_brush');

    if(this.brush_enabled === true) {
      // reset the brush's attributes
      this.brush_enabled = false;
      this.brush_from_edge = -1;
      this.brush_to_edge = -1;
      this.brush_from_name = -1;
      this.brush_to_name = -1;
      // remove the brush from CATview
      if (this.content)
        this.content.select(".brush").remove();
      // pass the full interval to the callback on remove
      if(this.brush_callback != null && this.edges.length > 0)
        this.brush_callback(this, 0, this.edges.length - 1, 0, this.names.length - 1);

    }
  };
  enable_drag = function(_mode){
    // if(this.debug)
    console.log('CATview.enable_drag');

    if(_mode === 'swap')
      this.drag_mode = 'swap'
    else
      this.drag_mode = 'insert'    // default

    this.drag_enabled = true;

    if(this.svg !== null && this.svg.select(".names-axis").size() === 1 ){
      this.svg.append("g").attr("class", "axis names-axis ");
      let axis = document.querySelector('#'+this.parent_id+' .names-axis');
      axis.classList.add('draggable');
      this.draw_names_axis();
    }

    return true;
  };
  disable_drag = function(){
    if(this.debug)
      console.log('CATview.disable_drag');

    this.drag_enabled = false;

    if(this.svg !== null && this.svg.select(".names-axis").size() === 1 ){
      this.svg.append("g").attr("class", "axis names-axis ");
      let axis = document.querySelector('#'+this.parent_id+' .names-axis');
      axis.classList.remove('draggable');
      this.draw_names_axis();
    }

    return true;
  };
};
